#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import enum
from asn1crypto import core

class SDFlagsRequest(enum.IntFlag):
	OWNER_SECURITY_INFORMATION = 0x1 #Owner identifier of the object.(OSI)
	GROUP_SECURITY_INFORMATION = 0x2 #Primary group identifier.(GSI)
	DACL_SECURITY_INFORMATION = 0x4 #Discretionary access control list (DACL) of the object.(DSI)
	SACL_SECURITY_INFORMATION = 0x8 #System access control list (SACL) of the object.(SSI) 

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/3888c2b7-35b9-45b7-afeb-b772aa932dd0
class SDFlagsRequestValue(core.Sequence):
	_fields = [
		('Flags', core.Integer),
	]




def calcualte_length(data):
	"""
	LDAP protocol doesnt send the total length of the message in the header,
	it only sends raw ASN1 encoded data structures, which has the length encoded.
	This function "decodes" the length os the asn1 structure, and returns it as int.
	"""
	if data[1] <= 127:
		return data[1] + 2
	else:
		bcount = data[1] - 128
		#if (bcount +2 ) > len(data):
		#	raise Exception('LDAP data too larage! Length byte count: %s' % bcount)
		return int.from_bytes(data[2:2+bcount], byteorder = 'big', signed = False) + bcount + 2




import os
from typing import Callable, Iterable, List, Optional

from prompt_toolkit.completion import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document
import traceback

class PathCompleter(Completer):
	"""
	Complete for Path variables.

	:param get_paths: Callable which returns a list of directories to look into
					  when the user enters a relative path.
	:param file_filter: Callable which takes a filename and returns whether
						this file should show up in the completion. ``None``
						when no filtering has to be done.
	:param min_input_len: Don't do autocompletion when the input string is shorter.
	"""
	def __init__(self, only_directories: bool = False,
				 get_current_dirs: Optional[Callable[[], List[str]]] = None,
				 file_filter: Optional[Callable[[str], bool]] = None,
				 min_input_len: int = 0,
				 expanduser: bool = False) -> None:

		self.only_directories = only_directories
		self.get_current_dirs = get_current_dirs or (lambda: ['.'])
		self.file_filter = file_filter or (lambda _: True)
		self.min_input_len = min_input_len
		self.expanduser = expanduser

	def get_completions(self, document: Document,
						complete_event: CompleteEvent) -> Iterable[Completion]:
		text = document.text_before_cursor

		# Complete only when we have at least the minimal input length,
		# otherwise, we can too many results and autocompletion will become too
		# heavy.
		if len(text) < self.min_input_len:
			return

		try:
			#print('Called!')
			# Do tilde expansion.
			#if self.expanduser:
			#	text = os.path.expanduser(text)

			# Directories where to look.
			dirnames = self.get_current_dirs()
			#print(text)
			for dirname in dirnames:
				if dirname.startswith(text):
					completion = dirname[len(text):]
					yield Completion(completion, 0, display=dirname)
				#elif dirname.find('=') != -1:
				#	m = dirname.find('=') + 1
				#	if dirname[m:].startswith(text):
				#		completion = dirname
				#		yield Completion(completion, 0, display=dirname)
		except Exception as e:
			traceback.print_exc()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

# https://tools.ietf.org/html/rfc4511
# https://msdn.microsoft.com/en-us/library/cc223501.aspx


from asn1crypto import core
import enum
import os

TAG = 'explicit'

# class
UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2

#https://msdn.microsoft.com/en-us/library/cc223359.aspx
class MSLDAPCapabilities(core.ObjectIdentifier):
	_map = {
		'1.2.840.113556.1.4.800' : 'LDAP_CAP_ACTIVE_DIRECTORY_OID',
		'1.2.840.113556.1.4.1791': 'LDAP_CAP_ACTIVE_DIRECTORY_LDAP_INTEG_OID',
		'1.2.840.113556.1.4.1670': 'LDAP_CAP_ACTIVE_DIRECTORY_V51_OID',
		'1.2.840.113556.1.4.1880': 'LDAP_CAP_ACTIVE_DIRECTORY_ADAM_DIGEST_OID',
		'1.2.840.113556.1.4.1851': 'LDAP_CAP_ACTIVE_DIRECTORY_ADAM_OID',
		'1.2.840.113556.1.4.1920': 'LDAP_CAP_ACTIVE_DIRECTORY_PARTIAL_SECRETS_OID',
		'1.2.840.113556.1.4.1935': 'LDAP_CAP_ACTIVE_DIRECTORY_V60_OID',
		'1.2.840.113556.1.4.2080': 'LDAP_CAP_ACTIVE_DIRECTORY_V61_R2_OID',
		'1.2.840.113556.1.4.2237': 'LDAP_CAP_ACTIVE_DIRECTORY_W8_OID',
	}

class scope(core.Enumerated):
	_map = {
		0 : 'baseObject',
		1 : 'singleLevel',
		2 : 'wholeSubtree',
	}

class derefAliases(core.Enumerated):
	_map = {
		0 : 'neverDerefAliases',
		1 : 'derefInSearching',
		2 : 'derefFindingBaseObj',
		3 : 'derefAlways',
	}

class resultCode(core.Enumerated):
	_map = {
		0  : 'success',
		1  : 'operationsError',
		2  : 'protocolError',
		3  : 'timeLimitExceeded',
		4  : 'sizeLimitExceeded',
		5  : 'compareFalse',
		6  : 'compareTrue',
		7  : 'authMethodNotSupported',
		8  : 'strongerAuthRequired',
		10 : 'referral',
		11 : 'adminLimitExceeded',
		12 : 'unavailableCriticalExtension',
		13 : 'confidentialityRequired',
		14 : 'saslBindInProgress',
		16 : 'noSuchAttribute',
		17 : 'undefinedAttributeType',
		18 : 'inappropriateMatching',
		19 : 'constraintViolation',
		20 : 'attributeOrValueExists',
		21 : 'invalidAttributeSyntax',
		32 : 'noSuchObject',
		33 : 'aliasProblem',
		34 : 'invalidDNSyntax',
		36 : 'aliasDereferencingProblem',
		48 : 'inappropriateAuthentication',
		49 : 'invalidCredentials',
		50 : 'insufficientAccessRights',
		51 : 'busy',
		52 : 'unavailable',
		53 : 'unwillingToPerform',
		54 : 'loopDetect',
		64 : 'namingViolation',
		65 : 'objectClassViolation',
		66 : 'notAllowedOnNonLeaf',
		67 : 'notAllowedOnRDN',
		68 : 'entryAlreadyExists',
		69 : 'objectClassModsProhibited',
		71 : 'affectsMultipleDSAs',
		80 : 'other',
	}

class ChangeOperation(core.Enumerated):
	_map = {
		0 : 'add',
		1 : 'delete',
		2 : 'replace',
	}	

class LDAPString(core.OctetString):
	pass
	
class LDAPDN(core.OctetString):
	pass
	
class LDAPOID(core.OctetString):
	pass
	
class URI(LDAPString):
	pass
	
class Referral(core.SequenceOf):
	_child_spec = URI

# https://www.ietf.org/rfc/rfc2696.txt
class SearchControlValue(core.Sequence):
	_fields = [
		('size' , core.Integer),
		('cookie' , core.OctetString)
	]

class Control(core.Sequence):
	_fields = [
		('controlType', LDAPOID),
		('criticality', core.Boolean, {'default' : False }),
		('controlValue', core.OctetString, {'optional': True })	,
	]
	
class Controls(core.SequenceOf):
	_child_spec = Control
	
class SaslCredentials(core.Sequence):
	_fields = [
		('mechanism', core.OctetString),
		('credentials', core.OctetString, {'optional': True}),	
	]
	
class SicilyPackageDiscovery(core.OctetString):
	pass

class SicilyNegotiate(core.OctetString):
	pass

class SicilyResponse(core.OctetString):
	pass
	
class AuthenticationChoice(core.Choice):
	_alternatives = [
		('simple', core.OctetString, {'implicit': (CONTEXT, 0)}),
		('sasl', SaslCredentials, {'implicit': (CONTEXT, 3)}),
		('sicily_disco', SicilyPackageDiscovery, {'implicit': (CONTEXT, 9)}), # FUCK
		('sicily_nego', SicilyNegotiate, {'implicit': (CONTEXT, 10)}), #YOU
		('sicily_resp', SicilyResponse, {'implicit': (CONTEXT, 11)}), #MICROSOFT
	]

class BindRequest(core.Sequence):
	_fields = [
		('version', core.Integer),
		('name', core.OctetString),
		('authentication', AuthenticationChoice),	
	]
	
class BindResponse(core.Sequence):
	_fields = [
		('resultCode', resultCode),
		('matchedDN', LDAPDN),
		('diagnosticMessage', LDAPString),
		('referral', Referral, {'optional': True}),
		('serverSaslCreds', core.OctetString, {'implicit': (CONTEXT, 7), 'optional': True}),
	]

class AttributeDescription(LDAPString):
	pass

class AttributeValue(core.OctetString):
	pass


class MatchingRuleId(LDAPString):
	pass

class AssertionValue(core.OctetString):
	pass

class AttributeValueAssertion(core.Sequence):
	_fields = [
		('attributeDesc', AttributeDescription),
		('assertionValue', AssertionValue),
	]

class Substring(core.Choice):
	_alternatives = [
		('initial', AssertionValue, {'implicit': (CONTEXT , 0) }  ),
		('any', AssertionValue, {'implicit': (CONTEXT , 1) }  ),
		('final', AssertionValue, {'implicit': (CONTEXT , 2) }  ),
	]
	

class Substrings(core.SequenceOf):
	_child_spec = Substring

class SubstringFilter(core.Sequence):
	_fields = [
		('type', AttributeDescription),
		('substrings', Substrings),
	]

class MatchingRuleAssertion(core.Sequence):
	_fields = [
		('matchingRule', MatchingRuleId, {'implicit': (CONTEXT, 1), 'optional' : True}  ),
		('type', AttributeDescription, {'implicit': (CONTEXT, 2), 'optional' : True}  ),
		('matchValue', AssertionValue, {'implicit': (CONTEXT, 3)}  ),
		('dnAttributes', core.Boolean, {'implicit': (CONTEXT, 4), 'default' : False}  ),
	]

# keep this Filter definition here! It is needed because filter class contains itself!
class Filter(core.Choice):
	pass

class Filters(core.SequenceOf):
	_child_spec = Filter

Filter._alternatives = [
		('and', Filters, {'implicit': (CONTEXT , 0) }  ),
		('or', Filters, {'implicit': (CONTEXT , 1) }  ),
		('not', Filter, {'explicit': (CONTEXT , 2) }  ), # https://tools.ietf.org/html/rfc4511#section-4.5.1.8
		('equalityMatch', AttributeValueAssertion, {'implicit': (CONTEXT , 3) }  ),
		('substrings', SubstringFilter, {'implicit': (CONTEXT , 4) }  ),
		('greaterOrEqual', AttributeValueAssertion, {'implicit': (CONTEXT , 5) }  ),
		('lessOrEqual', AttributeValueAssertion, {'implicit': (CONTEXT , 6) }  ),
		('present', AttributeDescription, {'implicit': (CONTEXT , 7) }  ),
		('approxMatch', AttributeValueAssertion, {'implicit': (CONTEXT , 8) }  ),
		('extensibleMatch', MatchingRuleAssertion, {'implicit': (CONTEXT , 9) }  ),

	]

class AttributeSelection(core.SequenceOf):
	_child_spec = LDAPString

class SearchRequest(core.Sequence):
	_fields = [
		('baseObject', LDAPDN),
		('scope', scope),
		('derefAliases', derefAliases),
		('sizeLimit', core.Integer),
		('timeLimit', core.Integer),
		('typesOnly', core.Boolean),
		('filter', Filter),
		('attributes', AttributeSelection),
	]

class AttributeValueSet(core.SetOf):
	_child_spec = AttributeValue


class PartialAttribute(core.Sequence):
	_fields = [
		('type', AttributeDescription),
		('attributes', AttributeValueSet),
	]

class PartialAttributeList(core.SequenceOf):
	_child_spec = PartialAttribute

class SearchResultEntry(core.Sequence):
	_fields = [
		('objectName', LDAPDN),
		('attributes', PartialAttributeList),
	]

class SearchResultReference(core.SequenceOf):
	_child_spec = URI

class UnbindRequest(core.Null):
	pass

class LDAPResult(core.Sequence):
	_fields = [
		('resultCode', resultCode ),
		('matchedDN', LDAPDN),
		('diagnosticMessage', LDAPString),
		('referral', Referral,  {'implicit': (CONTEXT, 3), 'optional': True}),
	]

class SearchResultDone(LDAPResult):
	pass

class Change(core.Sequence):
	_fields = [
		('operation', ChangeOperation),
		('modification', PartialAttribute),
	]

class Changes(core.SequenceOf):
	_child_spec = Change

class ModifyRequest(core.Sequence):
	_fields = [
		('object', LDAPDN),
		('changes', Changes),
	]

class ModifyResponse(LDAPResult):
	pass

class Attribute(PartialAttribute):
	pass

class AttributeList(core.SequenceOf):
	_child_spec = Attribute

class AddRequest(core.Sequence):
	_fields = [
		('entry', LDAPDN),
		('attributes', AttributeList),
	]

class AddResponse(LDAPResult):
	pass

class DelRequest(LDAPDN):
	pass

class DelResponse(LDAPResult):
	pass

class RelativeLDAPDN(LDAPString):
	pass

class ModifyDNRequest(core.Sequence):
	_fields = [
		('entry', LDAPDN),
		('newrdn', RelativeLDAPDN),
		('deleteoldrdn', core.Boolean),
		('newSuperior', LDAPDN, {'implicit': (CONTEXT, 0), 'optional': True}),
		
	]

class ModifyDNResponse(LDAPResult):
	pass

class CompareRequest(core.Sequence):
	_fields = [
		('entry', LDAPDN),
		('ava', AttributeValueAssertion),		
	]

class CompareResponse(LDAPResult):
	pass

class ExtendedRequest(core.Sequence):
	_fields = [
		('requestName', LDAPOID, {'implicit': (CONTEXT , 0) }),
		('requestValue', core.OctetString, {'implicit': (CONTEXT , 1), 'optional' : True}),		
	]

class ExtendedResponse(core.Sequence):
	_fields = [
		('resultCode', resultCode ),
		('matchedDN', LDAPDN),
		('diagnosticMessage', LDAPString),
		('referral', Referral,  {'implicit': (CONTEXT, 3), 'optional': True}),
		('responseName', LDAPOID, {'implicit': (CONTEXT , 10) , 'optional' : True}),
		('responseValue', core.OctetString, {'implicit': (CONTEXT , 11) , 'optional' : True}),
	]

class IntermediateResponse(core.Sequence):
	_fields = [
		('responseName', LDAPOID, {'implicit': (CONTEXT , 0) , 'optional' : True}),
		('responseValue', core.OctetString, {'implicit': (CONTEXT , 1) , 'optional' : True}),
	]

class protocolOp(core.Choice):
	_alternatives = [
		('bindRequest', BindRequest, {'implicit': (APPLICATION , 0) }  ),
		('bindResponse', BindResponse, {'implicit': (APPLICATION , 1) }  ),
		('unbindRequest', UnbindRequest, {'implicit': (APPLICATION,2) }  ),
		('searchRequest', SearchRequest, {'implicit': (APPLICATION,3) }  ),
		('searchResEntry', SearchResultEntry, {'implicit': (APPLICATION,4) }  ),
		('searchResDone', SearchResultDone, {'implicit': (APPLICATION,5) }  ),
		('modifyRequest', ModifyRequest, {'implicit': (APPLICATION,6) }  ),
		('modifyResponse', ModifyResponse, {'implicit': (APPLICATION,7) }  ),
		('addRequest', AddRequest, {'implicit': (APPLICATION,8) }  ),
		('addResponse', AddResponse, {'implicit': (APPLICATION,9) }  ),
		('delRequest', DelRequest, {'implicit': (APPLICATION,10) }  ),
		('delResponse', DelResponse, {'implicit': (APPLICATION,11) }  ),
		('modDNRequest', ModifyDNRequest, {'implicit': (APPLICATION,12) }  ),
		('modDNResponse', ModifyDNResponse, {'implicit': (APPLICATION,13) }  ),
		('compareRequest', CompareRequest, {'implicit': (APPLICATION,14) }  ),
		('compareResponse', CompareResponse, {'implicit': (APPLICATION,15) }  ),
		('abandonRequest', core.Integer, {'implicit': (APPLICATION,16) }  ), #integer is the messageid to abandon
		('searchResRef', SearchResultReference, {'implicit': (APPLICATION,19) }  ),
		('extendedReq', ExtendedRequest, {'implicit': (APPLICATION,23) }  ),
		('extendedResp', ExtendedResponse, {'implicit': (APPLICATION,24) }  ),
		('intermResponse', IntermediateResponse, {'implicit': (APPLICATION,25) }  ),
		
	]
	
class LDAPMessage(core.Sequence):
	_fields = [
		('messageID', core.Integer),
		('protocolOp', protocolOp),
		('controls', Controls, {'implicit': (CONTEXT, 0), 'optional': True}),	
	]
	



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import base64
from badldap.commons.utils import bh_dt_convert



MSADOU_ATTRS = [ 	
	'description', 'distinguishedName', 'dSCorePropagationData', 'gPLink', 'instanceType', 
	'isCriticalSystemObject', 'name', 'nTSecurityDescriptor', 'objectCategory', 'objectClass', 
	'objectGUID', 'ou', 'showInAdvancedViewOnly', 'systemFlags', 'uSNChanged', 'uSNCreated',
	'whenChanged', 'whenCreated', 'isDeleted'
]

class MSADOU:
	def __init__(self):
		self.description = None #dunno
		self.distinguishedName = None #datetime
		self.dSCorePropagationData = None #str
		self.gPLink = None #list
		self.instanceType = None #int
		self.isCriticalSystemObject = None #int
		self.name = None #int
		self.nTSecurityDescriptor = None #int
		self.objectCategory = None #int
		self.objectClass = None #str
		self.objectGUID = None #int
		self.ou = None #int
		self.showInAdvancedViewOnly = None #int
		self.systemFlags = None #str
		self.uSNChanged = None #int
		self.uSNCreated = None #str
		self.whenChanged = None #str
		self.whenCreated = None #str
		self.isDeleted = None #str
	
	@staticmethod
	def from_ldap(entry):
		adi = MSADOU()
		adi.description = entry['attributes'].get('description') #dunno
		adi.distinguishedName = entry['attributes'].get('distinguishedName') #datetime
		adi.dSCorePropagationData = entry['attributes'].get('dSCorePropagationData') #str
		adi.gPLink = entry['attributes'].get('gPLink') #list
		adi.instanceType = entry['attributes'].get('instanceType') #int
		adi.isCriticalSystemObject = entry['attributes'].get('isCriticalSystemObject') #int
		adi.name = entry['attributes'].get('name') #str
		adi.nTSecurityDescriptor = entry['attributes'].get('nTSecurityDescriptor') #str
		adi.objectCategory = entry['attributes'].get('objectCategory') #str
		adi.objectClass = entry['attributes'].get('objectClass') #str
		adi.objectGUID = entry['attributes'].get('objectGUID') #str
		adi.ou = entry['attributes'].get('ou') #str
		adi.showInAdvancedViewOnly = entry['attributes'].get('showInAdvancedViewOnly') #int
		adi.systemFlags = entry['attributes'].get('systemFlags') #int
		adi.uSNChanged = entry['attributes'].get('uSNChanged') #int
		adi.uSNCreated = entry['attributes'].get('uSNCreated') #int
		adi.whenChanged = entry['attributes'].get('whenChanged') #datetime
		adi.whenCreated = entry['attributes'].get('whenCreated') #datetime
		adi.isDeleted = entry['attributes'].get('isDeleted') #datetime
		return adi
	
	def to_dict(self):
		d = {}
		d['description'] = self.description
		d['distinguishedName'] = self.distinguishedName
		d['dSCorePropagationData'] = self.dSCorePropagationData
		d['gPLink'] = self.gPLink
		d['instanceType'] = self.instanceType
		d['isCriticalSystemObject'] = self.isCriticalSystemObject
		d['name'] = self.name
		d['nTSecurityDescriptor'] = self.nTSecurityDescriptor
		d['objectCategory'] = self.objectCategory
		d['objectClass'] = self.objectClass
		d['objectGUID'] = self.objectGUID
		d['ou'] = self.ou
		d['showInAdvancedViewOnly'] = self.showInAdvancedViewOnly
		d['systemFlags'] = self.systemFlags
		d['uSNChanged'] = self.uSNChanged
		d['uSNCreated'] = self.uSNCreated
		d['whenChanged'] = self.whenChanged
		d['whenCreated'] = self.whenCreated
		d['isDeleted'] = self.isDeleted
		return d

	def get_row(self, attrs):
		t = self.to_dict()
		if 'nTSecurityDescriptor' in attrs:
			if t['nTSecurityDescriptor'] is not None:
				t['nTSecurityDescriptor'] = base64.b64encode(t['nTSecurityDescriptor']).decode()
			else:
				t['nTSecurityDescriptor'] = b''
		return [str(t.get(x)) for x in attrs]

	def __str__(self):
		t = 'MSADOU\r\n'
		d = self.to_dict()
		for k in d:
			t += '%s: %s\r\n' % (k, d[k])
		return t 
	
	def to_bh(self, domain, domainsid):
		return {
			'Aces' : [],
			'Links'	: [],
			'ObjectIdentifier' : self.objectGUID.upper(),
			"IsDeleted": bool(self.isDeleted),
			"IsACLProtected": False , # Post processing
			'Properties' : {
				'name' : '%s@%s' % (self.name.upper(), domain.upper()),
				'domain' : domain,
				'domainsid' : domainsid, 
				'distinguishedname' : str(self.distinguishedName).upper(), 
				'highvalue' : False, # seems always false
				'whencreated' : bh_dt_convert(self.whenCreated),
				'description' : self.description ,				
				'blocksinheritance' : False, # seems always false
			},
			"GPOChanges": {
                "LocalAdmins": [],
                "RemoteDesktopUsers": [],
                "DcomUsers": [],
                "PSRemoteUsers": [],
                "AffectedComputers": []
            },
			'_gPLink' : self.gPLink,
		}


import io
from unicrypto.hashlib import md4

class MSDS_MANAGEDPASSWORD_BLOB:
	def __init__(self):
		self.Version = None
		self.Reserved = None
		self.Length = None
		self.CurrentPasswordOffset = None
		self.PreviousPasswordOffset = None
		self.QueryPasswordIntervalOffset = None
		self.UnchangedPasswordIntervalOffset = None
		self.CurrentPassword = None
		self.PreviousPassword = None
		#('AlignmentPadding',':'),
		self.QueryPasswordInterval = None
		self.UnchangedPasswordInterval = None

		self.nt_hash = None

	@staticmethod
	def from_bytes(data:bytes):
		return MSDS_MANAGEDPASSWORD_BLOB.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff:io.BytesIO):
		blob = MSDS_MANAGEDPASSWORD_BLOB()
		blob.Version = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		blob.Reserved = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		blob.Length = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.CurrentPasswordOffset = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		blob.PreviousPasswordOffset = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		blob.QueryPasswordIntervalOffset = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		blob.UnchangedPasswordIntervalOffset = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		
		ppwo = blob.PreviousPasswordOffset
		if ppwo == 0:
			ppwo = blob.QueryPasswordIntervalOffset
		if blob.CurrentPasswordOffset >0:
			buff.seek(blob.CurrentPasswordOffset, 0)
			blob.CurrentPassword = buff.read(ppwo - blob.CurrentPasswordOffset)
		if blob.PreviousPasswordOffset >0:
			buff.seek(blob.PreviousPasswordOffset, 0)
			blob.PreviousPassword = buff.read(blob.QueryPasswordIntervalOffset - blob.CurrentPasswordOffset)
		if blob.QueryPasswordIntervalOffset >0:
			buff.seek(blob.QueryPasswordIntervalOffset, 0)
			blob.QueryPasswordInterval = buff.read(blob.UnchangedPasswordIntervalOffset - blob.UnchangedPasswordIntervalOffset)
		if blob.UnchangedPasswordIntervalOffset >0:
			buff.seek(blob.UnchangedPasswordIntervalOffset, 0)
			blob.UnchangedPasswordInterval = buff.read()


		blob.nt_hash = md4(blob.CurrentPassword[:-2]).hexdigest()		
		return blob
	
	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
		return t


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import datetime #, timedelta, timezone
from badldap.ldap_objects.common import MSLDAP_UAC, vn

MSADDMSAUser_ATTRS = [ 	
	'accountExpires', 'badPasswordTime', 'badPwdCount', 'cn', 'codePage', 
	'distinguishedName', 'lastLogoff', 'lastLogon', 'lastLogonTimestamp', 
	'logonCount', 'name', 'objectCategory', 'objectClass', 'objectGUID', 
	'objectSid', 'primaryGroupID', 'pwdLastSet', 'sAMAccountName', 
	'sAMAccountType', 'sn', 'userAccountControl', 'whenChanged', 'whenCreated',
	'dNSHostName', 'msDS-SupportedEncryptionTypes',	'msDS-ManagedPasswordId', 
	'msDS-ManagedPasswordInterval', 'msDS-GroupMSAMembership', 'msDS-ManagedPassword',
	'msDS-ManagedAccountPrecededByLink', 'msDS-DelegatedMSAState'
]
MSADDMSAUser_TSV_ATTRS = [
	'sAMAccountName', 'badPasswordTime', 'badPwdCount', 'pwdLastSet', 'lastLogonTimestamp',
	'whenCreated', 'whenChanged', 'objectSid', 'cn', 'UAC_SCRIPT', 'UAC_ACCOUNTDISABLE', 
	'UAC_LOCKOUT', 'UAC_PASSWD_NOTREQD', 'UAC_PASSWD_CANT_CHANGE', 'UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED', 
	'UAC_DONT_EXPIRE_PASSWD', 'UAC_USE_DES_KEY_ONLY', 'UAC_DONT_REQUIRE_PREAUTH', 'UAC_PASSWORD_EXPIRED'
]

class MSADDMSAUser:
	def __init__(self):
		## ID
		self.sn = None #str
		self.cn = None #str
		self.distinguishedName = None #dn
		self.displayName = None #str
		self.name = None #str
		self.objectCategory = None #dn
		self.objectClass = None #str
		self.objectGUID = None #uid
		self.objectSid = None #str
		self.primaryGroupID = None #uid
		self.sAMAccountName = None #str
		self.dNSHostName = None #str
		self.msDS_SupportedEncryptionTypes = None #int
		self.msDS_ManagedPasswordId = None #bytes
		self.msDS_ManagedPasswordInterval = None #str
		self.msDS_GroupMSAMembership = None #SD
		self.msDS_ManagedPassword = None #str
		self.msDS_ManagedAccountPrecededByLink = None
		self.msDS_DelegatedMSAState = None

		## times
		self.accountExpires = None #datetime
		self.badPasswordTime = None #datetime
		self.lastLogoff = None #datetime
		self.lastLogon = None #datetime
		self.lastLogonTimestamp = None #datetime
		self.pwdLastSet = None #datetime
		self.whenChanged = None #datetime
		self.whenCreated = None #datetime

		## security
		self.badPwdCount = None #int
		self.logonCount = None #int
		self.sAMAccountType = None #int
		self.userAccountControl = None #UserAccountControl intflag
		
		## calculated properties
		self.when_pw_change = None #datetime
		self.when_pw_expires = None #datetime
		self.must_change_pw = None #datetime

	@staticmethod
	def from_ldap(entry, adinfo = None):
		adi = MSADDMSAUser()
		adi.sn = entry['attributes'].get('sn') 
		adi.cn = entry['attributes'].get('cn') 
		adi.distinguishedName = entry['attributes'].get('distinguishedName')
		adi.name = entry['attributes'].get('name')
		adi.objectCategory = entry['attributes'].get('objectCategory')
		adi.objectClass = entry['attributes'].get('objectClass')
		adi.objectGUID = entry['attributes'].get('objectGUID')
		adi.objectSid = entry['attributes'].get('objectSid')
		adi.primaryGroupID = entry['attributes'].get('primaryGroupID')
		adi.sAMAccountName = entry['attributes'].get('sAMAccountName')
		adi.accountExpires = entry['attributes'].get('accountExpires')
		adi.badPasswordTime = entry['attributes'].get('badPasswordTime')
		adi.lastLogoff = entry['attributes'].get('lastLogoff')
		adi.lastLogon = entry['attributes'].get('lastLogon')
		adi.lastLogonTimestamp = entry['attributes'].get('lastLogonTimestamp')
		adi.pwdLastSet = entry['attributes'].get('pwdLastSet')
		adi.whenChanged = entry['attributes'].get('whenChanged')
		adi.whenCreated = entry['attributes'].get('whenCreated')
		adi.badPwdCount = entry['attributes'].get('badPwdCount')
		adi.logonCount = entry['attributes'].get('logonCount')
		adi.sAMAccountType = entry['attributes'].get('sAMAccountType')
		adi.dNSHostName = entry['attributes'].get('dNSHostName')
		adi.msDS_SupportedEncryptionTypes = entry['attributes'].get('msDS-SupportedEncryptionTypes')
		adi.msDS_ManagedPasswordId = entry['attributes'].get('msDS-ManagedPasswordId')
		adi.msDS_ManagedPasswordInterval = entry['attributes'].get('msDS-ManagedPasswordInterval')
		adi.msDS_GroupMSAMembership = entry['attributes'].get('msDS-GroupMSAMembership')
		adi.msDS_ManagedPassword = entry['attributes'].get('msDS-ManagedPassword')
		adi.msDS_ManagedAccountPrecededByLink = entry['attributes'].get('msDS-ManagedAccountPrecededByLink')
		adi.msDS_DelegatedMSAState = entry['attributes'].get('msDS-DelegatedMSAState')
		
		temp = entry['attributes'].get('userAccountControl')
		if temp:
			adi.userAccountControl = MSLDAP_UAC(temp)
		return adi

	def to_dict(self):
		t = {}
		t['sn'] = vn(self.sn)
		t['cn'] = vn(self.cn)
		t['distinguishedName'] = vn(self.distinguishedName)
		t['name'] = vn(self.name)
		t['objectCategory'] = vn(self.objectCategory)
		t['objectClass'] = vn(self.objectClass)
		t['objectGUID'] = vn(self.objectGUID)
		t['objectSid'] = vn(self.objectSid)
		t['primaryGroupID'] = vn(self.primaryGroupID)
		t['sAMAccountName'] = vn(self.sAMAccountName)
		t['accountExpires'] = vn(self.accountExpires)
		t['badPasswordTime'] = vn(self.badPasswordTime)
		t['lastLogoff'] = vn(self.lastLogoff)
		t['lastLogon'] = vn(self.lastLogon)
		t['lastLogonTimestamp'] = vn(self.lastLogonTimestamp)
		t['pwdLastSet'] = vn(self.pwdLastSet)
		t['whenChanged'] = vn(self.whenChanged)
		t['whenCreated'] = vn(self.whenCreated)
		t['badPwdCount'] = vn(self.badPwdCount)
		t['logonCount'] = vn(self.logonCount)
		t['sAMAccountType'] = vn(self.sAMAccountType)
		t['userAccountControl'] = vn(self.userAccountControl)
		t['dNSHostName'] = vn(self.dNSHostName)
		t['msDS_SupportedEncryptionTypes'] = vn(self.msDS_SupportedEncryptionTypes)
		t['msDS_ManagedPasswordInterval'] = vn(self.msDS_ManagedPasswordInterval)
		t['msDS_GroupMSAMembership'] = vn(self.msDS_GroupMSAMembership)
		t['msDS_ManagedPassword'] = vn(self.msDS_ManagedPassword)
		t['msDS_ManagedPasswordId'] = vn(self.msDS_ManagedPasswordId)
		t['msDS_ManagedAccountPrecededByLink'] = vn(self.msDS_ManagedAccountPrecededByLink)
		t['msDS_DelegatedMSAState'] = vn(self.msDS_DelegatedMSAState)
		return t

	def uac_to_textflag(self, attr_s):
		if self.userAccountControl is None or self.userAccountControl == '':
			return 'N/A'
		attr = getattr(MSLDAP_UAC, attr_s[4:])
		if self.userAccountControl & attr:
			return True
		return False

	def get_row(self, attrs):
		t = self.to_dict()
		return [str(t.get(x)) if x[:4]!='UAC_' else str(self.uac_to_textflag(x)) for x in attrs]

	def __str__(self):
		t = 'MSADGMSAUser\n'
		t += 'sn: %s\n' % self.sn 
		t += 'cn: %s\n' % self.cn 
		t += 'distinguishedName: %s\n' % self.distinguishedName 
		t += 'name: %s\n' % self.name 
		t += 'primaryGroupID: %s\n' % self.primaryGroupID 
		t += 'sAMAccountName: %s\n' % self.sAMAccountName 
		t += 'accountExpires: %s\n' % self.accountExpires 
		t += 'badPasswordTime: %s\n' % self.badPasswordTime 
		t += 'lastLogoff: %s\n' % self.lastLogoff 
		t += 'lastLogon: %s\n' % self.lastLogon 
		t += 'lastLogonTimestamp: %s\n' % self.lastLogonTimestamp 
		t += 'pwdLastSet: %s\n' % self.pwdLastSet
		t += 'whenChanged: %s\n' % self.whenChanged
		t += 'whenCreated: %s\n' % self.whenCreated 
		t += 'objectGUID: %s\n' % self.objectGUID 
		t += 'objectSid: %s\n' % self.objectSid 
		t += 'badPwdCount: %s\n' % self.badPwdCount
		t += 'logonCount: %s\n' % self.logonCount
		t += 'sAMAccountType: %s\n' % self.sAMAccountType
		t += 'userAccountControl: %s\n' % self.userAccountControl
		t += 'dNSHostName %s\n' % self.dNSHostName
		t += 'msDS-SupportedEncryptionTypes: %s\n' % self.msDS_SupportedEncryptionTypes
		t += 'msDS-ManagedPasswordId: %s\n' % self.msDS_ManagedPasswordId
		t += 'msDS-ManagedPasswordInterval: %s\n' % self.msDS_ManagedPasswordInterval
		t += 'msDS-GroupMSAMembership: %s\n' % self.msDS_GroupMSAMembership
		t += 'msDS-ManagedPassword: %s\n' % self.msDS_ManagedPassword
		return t 



import datetime
from ipaddress import IPv4Address, IPv6Address
from typing import Union


def timestamp2datetime(dt: Union[bytes, int]) -> datetime.datetime:
    """
    Converting Windows timestamps to datetime.datetime format
    :param dt: Windows timestamp as array of bytes or integer
    :type dt: bytearray | int
    :return: datetime.datetime
    """

    if isinstance(dt, bytes):
        dt = int.from_bytes(dt, "little")

    return datetime.datetime(1601, 1, 1) + datetime.timedelta(microseconds=dt / 10)


def datetime2timestamp(dt) -> int:
    delta = dt - datetime.datetime(1601, 1, 1)
    ns = int((delta / datetime.timedelta(microseconds=1)) * 10)
    return ns.to_bytes(8, "little", signed=False)


def bytes2ipv4(data: bytes) -> str:
    return str(IPv4Address(data))


def bytes2ipv6(data: bytes) -> str:
    return str(IPv6Address(data))


def wrap(s, w) -> str:
    return [s[i : i + w] for i in range(0, len(s), w)]


def print_cert(cert, offset=0) -> str:
    cert = cert["tbs_certificate"]
    blanks = " " * offset
    msg = [
        "Cert Subject: %s" % cert["subject"]["common_name"],
        "Cert Serial: %s" % cert["serial_number"],
        "Cert Start: %s" % cert["validity"]["not_before"],
        "Cert End: %s" % cert["validity"]["not_after"],
        "Cert Issuer: %s" % cert["issuer"]["common_name"],
    ]
    return "{}{}".format(blanks, "\n{}".format(blanks).join(msg))


def win_timestamp_to_unix(seconds):
    """
    Convert Windows timestamp (100 ns since 1 Jan 1601) to
    unix timestamp.
    """
    seconds = int(seconds)
    if seconds == 0:
        return 0
    return int((seconds - 116444736000000000) / 10000000)


def bh_dt_convert(dt: datetime.datetime):
    if dt is None or dt == 0 or dt == "0" or dt == "":
        return -1
    ts = max(0, int(dt.timestamp()))
    return ts


FUNCTIONAL_LEVELS = {
    0: "2000 Mixed/Native",
    1: "2003 Interim",
    2: "2003",
    3: "2008",
    4: "2008 R2",
    5: "2012",
    6: "2012 R2",
    7: "2016",
}

KNOWN_SIDS = {
    "S-1-0": "Null Authority",
    "S-1-0-0": "Nobody",
    "S-1-1": "World Authority",
    "S-1-1-0": "Everyone",
    "S-1-2": "Local Authority",
    "S-1-2-0": "Local",
    "S-1-3": "Creator Authority",
    "S-1-3-0": "Creator Owner",
    "S-1-3-1": "Creator Group",
    "S-1-3-4": "Owner Rights",
    "S-1-4": "Non-unique Authority",
    "S-1-5": "NT Authority",
    "S-1-5-1": "Dialup",
    "S-1-5-2": "Network",
    "S-1-5-3": "Batch",
    "S-1-5-4": "Interactive",
    "S-1-5-5-X-Y": "Logon Session",
    "S-1-5-6": "Service",
    "S-1-5-7": "Anonymous",
    "S-1-5-9": "Enterprise Domain Controllers",
    "S-1-5-10": "Principal Self",
    "S-1-5-11": "Authenticated Users",
    "S-1-5-12": "Restricted Code",
    "S-1-5-13": "Terminal Server Users",
    "S-1-5-14": "Remote Interactive Logon",
    "S-1-5-17": "IUSR",
    "S-1-5-18": "Local System",
    "S-1-5-19": "NT Authority Local Service",
    "S-1-5-20": "NT Authority Network Service",
    "S-1-5-32-544": "Administrators",
    "S-1-5-32-545": "Users",
    "S-1-5-32-546": "Guests",
    "S-1-5-32-547": "Power Users",
    "S-1-5-32-548": "Account Operators",
    "S-1-5-32-549": "Server Operators",
    "S-1-5-32-550": "Print Operators",
    "S-1-5-32-551": "Backup Operators",
    "S-1-5-32-552": "Replicators",
    "S-1-5-32-582": "Storage Replica Administrators",
    "S-1-5-64-10": "NTLM Authentication",
    "S-1-5-64-14": "SChannel Authentication",
    "S-1-5-64-21": "Digest Authentication",
    "S-1-5-80": "NT Service",
}




from badldap.protocol.messages import resultCode
from badldap.wintypes.winerror import WINERROR
import re


LDAPResultCodeLookup ={
	0  : 'success',
	1  : 'operationsError',
	2  : 'protocolError',
	3  : 'timeLimitExceeded',
	4  : 'sizeLimitExceeded',
	5  : 'compareFalse',
	6  : 'compareTrue',
	7  : 'authMethodNotSupported',
	8  : 'strongerAuthRequired',
	10 : 'referral',
	11 : 'adminLimitExceeded',
	12 : 'unavailableCriticalExtension',
	13 : 'confidentialityRequired',
	14 : 'saslBindInProgress',
	16 : 'noSuchAttribute',
	17 : 'undefinedAttributeType',
	18 : 'inappropriateMatching',
	19 : 'constraintViolation',
	20 : 'attributeOrValueExists',
	21 : 'invalidAttributeSyntax',
	32 : 'noSuchObject',
	33 : 'aliasProblem',
	34 : 'invalidDNSyntax',
	36 : 'aliasDereferencingProblem',
	48 : 'inappropriateAuthentication',
	49 : 'invalidCredentials',
	50 : 'insufficientAccessRights',
	51 : 'busy',
	52 : 'unavailable',
	53 : 'unwillingToPerform',
	54 : 'loopDetect',
	64 : 'namingViolation',
	65 : 'objectClassViolation',
	66 : 'notAllowedOnNonLeaf',
	67 : 'notAllowedOnRDN',
	68 : 'entryAlreadyExists',
	69 : 'objectClassModsProhibited',
	71 : 'affectsMultipleDSAs',
	80 : 'other',
}
LDAPResultCodeLookup_inv = {v: k for k, v in LDAPResultCodeLookup.items()}

class LDAPServerException(Exception):
	def __init__(self, resultname, diagnostic_message, message = None):
		self.resultcode = LDAPResultCodeLookup_inv[resultname]
		self.resultname = resultname
		self.diagnostic_message = diagnostic_message
		self.message = format_ad_ldap_error(diagnostic_message, self.resultname, getattr(self, "dn", None))
		super().__init__(self.message)

class LDAPSearchException(LDAPServerException):
	def __init__(self, resultcode, diagnostic_message):
		message = 'LDAP Search failed! Result code: "%s" Reason: "%s"' % (resultcode, diagnostic_message)
		super().__init__(resultcode, diagnostic_message, message)

class LDAPBindException(LDAPServerException):
	def __init__(self, resultcode, diagnostic_message):
		message = 'LDAP Bind failed! Result code: "%s" Reason: "%s"' % (resultcode, diagnostic_message)
		super().__init__(resultcode, diagnostic_message, message)

class LDAPAddException(LDAPServerException):
	def __init__(self, dn, resultcode, diagnostic_message):
		self.dn = dn
		message = 'LDAP Add operation failed on DN %s! Result code: "%s" Reason: "%s"' % (self.dn, resultcode, diagnostic_message)
		super().__init__(resultcode, diagnostic_message, message)

class LDAPModifyException(LDAPServerException):
	def __init__(self, dn, resultcode, diagnostic_message):
		self.dn = dn
		message = 'LDAP Modify operation failed on DN %s! Result code: "%s" Reason: "%s"' % (self.dn, resultcode, diagnostic_message)
		super().__init__(resultcode, diagnostic_message, message)

class LDAPDeleteException(LDAPServerException):
	def __init__(self, dn, resultcode, diagnostic_message):
		self.dn = dn
		message = 'LDAP Delete operation failed on DN %s! Result code: "%s" Reason: "%s"' % (self.dn, resultcode, diagnostic_message)
		super().__init__(resultcode, diagnostic_message, message)

class LDAPModifyDNException(LDAPServerException):
	def __init__(self, dn, resultcode, diagnostic_message):
		self.dn = dn
		message = 'LDAP ModifyDN operation failed on DN %s! Result code: "%s" Reason: "%s"' % (self.dn, resultcode, diagnostic_message)
		super().__init__(resultcode, diagnostic_message, message)


def format_ad_ldap_error(diag_raw: bytes, resultcode: str, dn: str = None) -> str:
	diag_msg = diag_raw.decode('utf-8', errors='ignore')

	object_error = ''
	if dn:
		# 	# If LDAP error on DN there is probably an attribute specified to extract
		m_attr = re.search(r'Att\s+[0-9A-Fa-f]+\s+\(([^)]+)\)', diag_msg)
		attribute = f' {m_attr.group(1)}' if m_attr else ''
		object_error = f"for {dn} (Attr{attribute}) "

	# Find first matching WINERROR code
	code_name = None
	code_message = None
	for hexstr in re.findall(r'\b[0-9A-Fa-f]{8}\b', diag_msg):
		try:
			code_int = int(hexstr, 16)
		except ValueError:
			continue
		info = WINERROR.get(code_int)
		if info:
			code_name = info.get("code")
			code_message = info.get("message")
			break

	reason = None
	if code_name:
		reason = f"({code_name}) {code_message}"
	else:
		m_reason = re.search(r'(problem[^,]+),\s+data 0([^\n]*)', diag_msg)
		raw_reason = ''.join(m_reason.groups()) if m_reason else None
		reason = raw_reason if raw_reason else diag_msg

	return f"{resultcode} {object_error}— Reason:{reason}"







def reverse_dn_components(dn:str):
	rdns = ','.join(reversed(dn.split(',')))
	return rdns.upper()

def explode_dn(dn):
	parts = []
	esc = False
	part = ''

	for char in dn:
		if esc:
			part += char
			esc = False
		elif char == '\\':
			esc = True
			part += char
		elif char == ',':
			if part:
				parts.append(part)
				part = ''
		else:
			part += char

	if part:
		parts.append(part)

	return parts


def parse_gplink_string(linkstr):
	if not linkstr:
		return
	for links in linkstr.split('[LDAP://')[1:]:
		dn, options = links.rstrip('][').split(';')
		yield dn, int(options)



#taken from bloodhound.py
def is_filtered_container(containerdn):
	if "CN=DOMAINUPDATES,CN=SYSTEM,DC=" in containerdn.upper():
		return True
	if "CN=POLICIES,CN=SYSTEM,DC=" in containerdn.upper() and (containerdn.upper().startswith('CN=USER') or containerdn.upper().startswith('CN=MACHINE')):
		return True
	return False

def is_filtered_container_child(containerdn):
	if "CN=PROGRAM DATA,DC=" in containerdn.upper():
		return True
	if "CN=SYSTEM,DC=" in containerdn.upper():
		return True
	return False